/*
 *  Copyright (c) 2008 Tomasz Moń <desowin@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <libmirage/mirage-error.h>
#include "icon.h"
#include "marshal.h"

#define LIBMIRAGE_ERROR_QUARK_CODE "org.freedesktop.DBus.GLib.UnmappedError.MirageErrorQuark.Code"

static DBusGConnection *connection = NULL;
static DBusGProxy *g_proxy = NULL;
static GtkStatusIcon *tray_icon;

typedef struct {
    GtkWidget *dpm_emu;
    GtkWidget *tr_emu;

    gulong dpm_emu_id; /* id of toggled signal handler */
    gulong tr_emu_id;  /* id of toggled signal handler */
} drive_menu;

static drive_menu* drive_menus = NULL;

static void update_tray_icon_tooltip();

static gboolean get_status(gint device, gboolean *loaded, gchar ***filenames) {
    GError *err = NULL;
    dbus_g_proxy_call(g_proxy, "DeviceGetStatus", &err,
                      G_TYPE_INT, device, G_TYPE_INVALID, G_TYPE_BOOLEAN, loaded, G_TYPE_STRV, filenames, G_TYPE_INVALID);

    if (err) {
        printf("Error calling DeviceGetStatus: %s\n", err->message);
        g_error_free(err);
        *loaded = FALSE;
        *filenames = NULL;
        g_error_free(err);
        return FALSE;
    }
    return TRUE;
}

void device_get_maping(gint device, gchar **sr_device, gchar **sg_device) {
    GError *err = NULL;
    dbus_g_proxy_call(g_proxy, "DeviceGetMapping", &err,
                      G_TYPE_INT, device, G_TYPE_INVALID, G_TYPE_STRING, sr_device, G_TYPE_STRING, sg_device, G_TYPE_INVALID);

    if (err) {
        printf("Error calling DeviceGetMapping: %s\n", err->message);
        *sr_device = *sg_device = NULL;
        g_error_free(err);
    }
}

static gint get_number_of_devices() {
    GError *err = NULL;
    gint no = 0;
    dbus_g_proxy_call(g_proxy, "GetNumberOfDevices", &err,
                      G_TYPE_INVALID, G_TYPE_INT, &no, G_TYPE_INVALID);
    if (err) {
        printf("Error calling GetNumberOfDevices: %s\n", err->message);
        g_error_free(err);
        return 0;
    }
    return no;
}

static GPtrArray *get_parsers() {
    GError *err = NULL;
    GPtrArray *parsers = NULL;
    dbus_g_proxy_call(g_proxy, "EnumSupportedParsers", &err,
                      G_TYPE_INVALID,
                      dbus_g_type_get_collection("GPtrArray", G_TYPE_VALUE_ARRAY), &parsers, G_TYPE_INVALID);
    if (err) {
        printf("Error calling EnumSupportedParsers: %s\n", err->message);
        g_error_free(err);
        return NULL;
    }

    printf("Supported parsers:\n");
    int x;
    for (x=0; x<parsers->len; x++) {
        GValueArray *array = g_ptr_array_index(parsers, x);
        int y;
        for (y=0; y<array->n_values; y++) {
            GValue *value = g_value_array_get_nth(array, y);
            if (G_VALUE_TYPE(value) == G_TYPE_STRING)
                printf("%s ", g_value_get_string(value));
        }
        printf("\n");
    }

    return parsers;
}

static gboolean set_options(gint device, const gchar *option, GPtrArray *options) {
    GError *err = NULL;
    dbus_g_proxy_call(g_proxy, "DeviceSetOption", &err,
                      G_TYPE_INT, device, G_TYPE_STRING, option, dbus_g_type_get_collection("GPtrArray", G_TYPE_VALUE), options,
                      G_TYPE_INVALID, G_TYPE_INVALID);
    if (err) {
        printf("Error calling DeviceSetOption: %s\n", err->message);
        g_error_free(err);
        return FALSE;
    }
    return TRUE;
}

static gboolean set_boolean_option(gint device, gboolean value, const gchar *option) {
    GPtrArray *options = g_ptr_array_new();
    GValue *val = g_new0(GValue,1);
    g_value_init(val, G_TYPE_BOOLEAN);
    g_value_set_boolean(val, value);
    g_ptr_array_add(options, val);
    gboolean ret = set_options(device, option, options);
    g_ptr_array_free(options, TRUE);
    return ret;
}

static GPtrArray *get_options(gint device, const gchar *option) {
    GError *err = NULL;
    GPtrArray *options = NULL;
    dbus_g_proxy_call(g_proxy, "DeviceGetOption", &err,
                      G_TYPE_INT, device, G_TYPE_STRING, option, G_TYPE_INVALID,
                      dbus_g_type_get_collection("GPtrArray", G_TYPE_VALUE), &options, G_TYPE_INVALID);
    if (err) {
        printf("Error calling DeviceGetOption: %s\n", err->message);
        g_error_free(err);
        return NULL;
    }
    return options;
}

/* valid option are: "dpm-emulation", "tr-emulation" */
static gboolean get_boolean_option(gint device, gboolean *value, const gchar *option) {
     GPtrArray *options = get_options(device, option);
     if (options) {
         *value = g_value_get_boolean(g_ptr_array_index(options, 0));
         printf("Option %s = %d\n", option, *value);
         return TRUE;
     }
     return FALSE;
}

static gboolean get_device_id(gint device, gchar **vendor_id, gchar **product_id, gchar **revision, gchar **vendor_specific) {
    GPtrArray *options = get_options(device, "device-id");
    if (!options || options->len != 4)
        return FALSE;

    *vendor_id = g_strdup(g_value_get_string(g_ptr_array_index(options, 0)));
    *product_id = g_strdup(g_value_get_string(g_ptr_array_index(options, 1)));
    *revision = g_strdup(g_value_get_string(g_ptr_array_index(options, 2)));
    *vendor_specific = g_strdup(g_value_get_string(g_ptr_array_index(options, 3)));

    return TRUE;
}

static gboolean device_load(gint device, gchar *filename) {
    GError *err = NULL;
    /* UGLY AS HELL */
    gchar *filenames[2];
    filenames[0] = g_strdup(filename);
    filenames[1] = NULL;

    GHashTable *parameters = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, g_free);

    dbus_g_proxy_call(g_proxy, "DeviceLoad", &err,
                      G_TYPE_INT, device, G_TYPE_STRV, &filenames,
                      dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), parameters,
                      G_TYPE_INVALID, G_TYPE_INVALID);

    g_free(filenames[0]);
    g_hash_table_destroy(parameters);

    if (err) {
        if (err->domain == DBUS_GERROR && err->code == DBUS_GERROR_REMOTE_EXCEPTION) {
            const gchar *error_name = dbus_g_error_get_name(err);
            if (g_str_has_prefix(error_name, LIBMIRAGE_ERROR_QUARK_CODE)) {
                gint a = atoi(error_name + strlen(LIBMIRAGE_ERROR_QUARK_CODE));

                switch (a) {
                    case MIRAGE_E_NEEDPASSWORD:
                        printf("Need Password\n");
                        break;
                    case MIRAGE_E_WRONGPASSWORD:
                        printf("Wrong Password\n");
                        break;
                }
            }
        }

        GtkWidget *msg_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL,
                                                       GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
                                                       "Error calling DeviceLoad on device %d:\n%s",
                                                       device, err->message);
        gtk_dialog_run(GTK_DIALOG(msg_dialog));
        gtk_widget_destroy(msg_dialog);
        g_error_free(err);
        return FALSE;
    }

    update_tray_icon_tooltip();
    return TRUE;
}

static void device_unload(gint device) {
    GError *err = NULL;
    dbus_g_proxy_call(g_proxy, "DeviceUnload", &err,
                      G_TYPE_INT, device, G_TYPE_INVALID, G_TYPE_INVALID);

    if (err) {
        GtkWidget *msg_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL,
                                                       GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
                                                       "Error calling DeviceUnload on device %d:\n%s",
                                                       device, err->message);
        gtk_dialog_run(GTK_DIALOG(msg_dialog));
        gtk_widget_destroy(msg_dialog);
        g_error_free(err);
    }
    update_tray_icon_tooltip();
}

static gchar *devices_state() {
    int i;
    gboolean loaded;
    gchar **filenames;

    gint devices_num = get_number_of_devices();
    GString *tmp = g_string_new("DEV LOADED FILENAME\n");

    for (i = 0; i<devices_num; i++) {
        if (!get_status(i, &loaded, &filenames))
            continue;

        g_string_append_printf(tmp, "%-4d%-7d", i, loaded);
        guint x;
        for (x = 0; x<g_strv_length(filenames); x++) {
            if (x == 0) {
                g_string_append_printf(tmp, "%s\n", filenames[x]);
            } else {
                g_string_append_printf(tmp, "%22s%s\n", "", filenames[x]);
            }
        }

        g_strfreev(filenames); filenames = NULL;
    }

    return g_string_free(tmp, FALSE);
}

static gchar *devices_maping() {
    gchar *sr_device, *sg_device;

    gint i;
    gint devices_num = get_number_of_devices();

    GString *tmp = g_string_new("DEV SCSI CD-ROM   SCSI generic\n");
    for (i = 0; i<devices_num; i++) {
         device_get_maping(i, &sr_device, &sg_device);
         g_string_append_printf(tmp, "%-4d%-14s%s\n", i, sr_device, sg_device);
    }

    return g_string_free(tmp, FALSE);
}

static void tray_icon_menu(GtkStatusIcon *status_icon, guint button, guint activate_time, GtkWidget *menu) {
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, button, activate_time);
}

static void popup_help_cb(GtkStatusIcon *status_icon, gpointer user_data) {
    gchar *maping = devices_maping();
    GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, maping, NULL);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
    g_free(maping);
}

static void load_cb(GtkMenuItem *status_icon, gpointer user_data) {
    gint device = GPOINTER_TO_INT(user_data);

    static gchar *last_directory = NULL;
    GtkWidget *dialog = gtk_file_chooser_dialog_new("Open Image", NULL,
				                    GTK_FILE_CHOOSER_ACTION_OPEN,
				                    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				                    GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				                    NULL);

    if (last_directory)
        gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), last_directory);

    GPtrArray *parsers = get_parsers();
    if (parsers) {
        GtkFileFilter *filter = gtk_file_filter_new();
        gtk_file_filter_set_name(filter, "All Supported Images");
        gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

        int x;
        for (x=0; x<parsers->len; x++) {
            GValueArray *array = g_ptr_array_index(parsers, x);
            GValue *value = NULL;

            GtkFileFilter *current_filter = gtk_file_filter_new();

            value = g_value_array_get_nth(array, 2);
            if (G_VALUE_TYPE(value) == G_TYPE_STRING)
                gtk_file_filter_set_name(current_filter, g_value_get_string(value));

            value = g_value_array_get_nth(array, 3);
            if (G_VALUE_TYPE(value) == G_TYPE_STRING) {
                gtk_file_filter_add_mime_type(current_filter, g_value_get_string(value));
                gtk_file_filter_add_mime_type(filter, g_value_get_string(value));
            }
            gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), current_filter);
        }
    }

    gboolean loaded = FALSE;
    while (!loaded && gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
        char *filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
        loaded = device_load(device, filename);
        g_free(filename);
        if (last_directory) g_free(last_directory);
        last_directory = g_strdup(gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(dialog)));
    }

    gtk_widget_destroy(dialog);
}

static void unload_cb(GtkMenuItem *status_icon, gpointer user_data) {
    gint device = GPOINTER_TO_INT(user_data);
    device_unload(device);
}

static void update_tray_icon_tooltip() {
    int i;
    gboolean loaded;
    gchar **filenames;
    gchar *sr, *sg;

    gint devices_num = get_number_of_devices();
    GString *tmp = g_string_new("Devices:");

    for (i = 0; i<devices_num; i++) {
        if (!get_status(i, &loaded, &filenames))
            continue;

        device_get_maping(i, &sr, &sg);
        g_string_append_printf(tmp, "\n%s: ", sr);
        g_free(sr);
        g_free(sg);
        if (loaded) {
            guint x;
            guint len = g_strv_length(filenames);
            for (x = 0; x<len; x++) {
                g_string_append_printf(tmp, (x==len-1) ? "%s" : "%s\n", filenames[x]);
            }
        } else
            g_string_append(tmp, "Empty");

        g_strfreev(filenames); filenames = NULL;
    }

    gtk_status_icon_set_tooltip(tray_icon, tmp->str);
    g_string_free(tmp, TRUE);
}

static void dpm_emulation_toggle_cb(GtkCheckMenuItem *item, gpointer data) {
    gint device = GPOINTER_TO_INT(data);
    gboolean state;
    g_object_get(G_OBJECT(item), "active", &state, NULL);
    printf("dpm emulation toggled to %d on device %d\n", state, device);
    set_boolean_option(device, state, "dpm-emulation");
}

static void tr_emulation_toggle_cb(GtkCheckMenuItem *item, gpointer data) {
    gint device = GPOINTER_TO_INT(data);
    gboolean state;
    g_object_get(G_OBJECT(item), "active", &state, NULL);
    printf("tr emulation toggled to %d on device %d\n", state, device);
    set_boolean_option(device, state, "tr-emulation");
}

static void setup_tray() {
    GdkPixbuf *pixbuf = gdk_pixbuf_new_from_inline (-1, tray_pixbuf, FALSE, NULL);
    gtk_window_set_default_icon(pixbuf);
    tray_icon = gtk_status_icon_new_from_pixbuf(pixbuf);
    g_object_unref(pixbuf);

    GtkWidget *menu, *menu_item;
    menu = gtk_menu_new();
    menu_item = gtk_menu_item_new_with_label("Help");
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
    g_signal_connect(menu_item, "activate", G_CALLBACK(popup_help_cb), NULL);

    gint x, devices;
    devices = get_number_of_devices();

    drive_menus = g_new0(drive_menu, devices);
    for (x = 0; x<devices; x++) {
         gchar *sr, *sg;
         device_get_maping(x, &sr, &sg);
         gchar *label = g_strdup_printf("%s", sr);
         g_free(sr);
         g_free(sg);
         menu_item = gtk_menu_item_new_with_label(label);
         gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
         g_free(label);

         GtkWidget *sub_menu, *sub_menu_item;
         sub_menu = gtk_menu_new();
         sub_menu_item = gtk_menu_item_new_with_label("Load");
         gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), sub_menu_item);
         g_signal_connect(sub_menu_item, "activate", G_CALLBACK(load_cb), GINT_TO_POINTER(x));

         sub_menu_item = gtk_menu_item_new_with_label("Unload");
         gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), sub_menu_item);
         g_signal_connect(sub_menu_item, "activate", G_CALLBACK(unload_cb), GINT_TO_POINTER(x));

         sub_menu_item = gtk_separator_menu_item_new();
         gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), sub_menu_item);

         gboolean active;
         sub_menu_item = gtk_check_menu_item_new_with_label("dpm-emulation");
         if (get_boolean_option(x, &active, "dpm-emulation"))
             gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(sub_menu_item), active);
         gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), sub_menu_item);
         drive_menus[x].dpm_emu_id = g_signal_connect(sub_menu_item, "toggled", G_CALLBACK(dpm_emulation_toggle_cb), GINT_TO_POINTER(x));
         drive_menus[x].dpm_emu = sub_menu_item;

         sub_menu_item = gtk_check_menu_item_new_with_label("tr-emulation");
         if (get_boolean_option(x, &active, "tr-emulation"))
             gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(sub_menu_item), active);
         gtk_menu_shell_append(GTK_MENU_SHELL(sub_menu), sub_menu_item);
         drive_menus[x].tr_emu_id = g_signal_connect(sub_menu_item, "toggled", G_CALLBACK(tr_emulation_toggle_cb), GINT_TO_POINTER(x));
         drive_menus[x].tr_emu = sub_menu_item;

         gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), sub_menu);
    }

    menu_item = gtk_menu_item_new_with_label("Quit");
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
    g_signal_connect(menu_item, "activate", G_CALLBACK(gtk_main_quit), NULL);
    gtk_widget_show_all(menu);

    update_tray_icon_tooltip();
    g_signal_connect(G_OBJECT(tray_icon), "popup-menu", G_CALLBACK(tray_icon_menu), menu);
}

static void device_option_changed_cb(DBusGProxy *proxy, gint device, const gchar *option, gpointer data) {
    printf("Device %d option %s has changed\n", device, option);
    gboolean value;
    if (g_strcmp0(option, "dpm-emulation") == 0) {
        if (get_boolean_option(device, &value, option)) {
            g_signal_handler_block(drive_menus[device].dpm_emu, drive_menus[device].dpm_emu_id);
            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(drive_menus[device].dpm_emu), value);
            g_signal_handler_unblock(drive_menus[device].dpm_emu, drive_menus[device].dpm_emu_id);
        }
        return;
    }
    if (g_strcmp0(option, "tr-emulation") == 0) {
        if (get_boolean_option(device, &value, option)) {
            g_signal_handler_block(drive_menus[device].tr_emu, drive_menus[device].tr_emu_id);
            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(drive_menus[device].tr_emu), value);
            g_signal_handler_unblock(drive_menus[device].tr_emu, drive_menus[device].tr_emu_id);
        }
        return;
    }
}

static void device_status_changed_cb(DBusGProxy *proxy, gint device, gpointer data) {
    printf("Device %d status changed\n", device);
    update_tray_icon_tooltip();
}

static void connect_signals() {
    dbus_g_object_register_marshaller(marshal_VOID__INT_STRING, G_TYPE_NONE, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal(g_proxy, "DeviceOptionChanged", G_TYPE_INT, G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(g_proxy, "DeviceOptionChanged", G_CALLBACK(device_option_changed_cb), NULL, NULL);

    dbus_g_proxy_add_signal(g_proxy, "DeviceStatusChanged", G_TYPE_INT, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(g_proxy, "DeviceStatusChanged", G_CALLBACK(device_status_changed_cb), NULL, NULL);
}

static void disconnect_signals() {
    dbus_g_proxy_disconnect_signal(g_proxy, "DeviceOptionChanged", G_CALLBACK(device_option_changed_cb), NULL);
    dbus_g_proxy_disconnect_signal(g_proxy, "DeviceStatusChanged", G_CALLBACK(device_status_changed_cb), NULL);
}

static gboolean check_for_cdemud(DBusGConnection *connection) {
    GError *err = NULL;
    DBusGProxy *g_proxy = dbus_g_proxy_new_for_name(connection,
                                                    "org.freedesktop.DBus",
                                                    "/",
                                                    "org.freedesktop.DBus");

    GStrv names = NULL;

    dbus_g_proxy_call(g_proxy, "ListNames", &err, G_TYPE_INVALID,
                      G_TYPE_STRV, &names, G_TYPE_INVALID);

    if (err) {
        printf("%s\n", err->message);
        return FALSE;
    }

    guint names_len = g_strv_length(names);
    guint i;
    for (i = 0; i < names_len; i++) {
         if (g_strcmp0("net.sf.cdemu.CDEMUD_Daemon", names[i]) == 0) {
             g_object_unref(g_proxy);
             return TRUE;
         }
    }
    return FALSE;
}

int main(int argc, char **argv) {
    gtk_init(&argc, &argv);

    GError *err = NULL;
    connection = dbus_g_bus_get(DBUS_BUS_SYSTEM, &err);
    if (err) {
        printf("Error: %s\n", err->message);
        printf("Please check if dbus is working correctly\n");
        g_error_free(err);
        return 1;
    }

    if (!check_for_cdemud(connection)) {
        dbus_g_connection_unref(connection);
        err = NULL;

        connection = dbus_g_bus_get(DBUS_BUS_SESSION, &err);
        if (err) {
            printf("Error: %s\n", err->message);
            printf("Please check if dbus is working correctly\n");
            g_error_free(err);
            return 1;
        }
        if (!check_for_cdemud(connection)) {
            printf("No cdemud interface found neither on system or session bus\n");
            printf("Try 'sudo modprobe vhba && sudo cdemud -d'\n");
            dbus_g_connection_unref(connection);
            return 1;
        }
    }

    g_proxy = dbus_g_proxy_new_for_name(connection,
                                        "net.sf.cdemu.CDEMUD_Daemon",
                                        "/CDEMUD_Daemon",
                                        "net.sf.cdemu.CDEMUD_Daemon");

    gchar *version;
    err = NULL;
    dbus_g_proxy_call(g_proxy, "GetLibraryVersion", &err,
                      G_TYPE_INVALID, G_TYPE_STRING, &version, G_TYPE_INVALID);
    if (err) {
        printf("Error calling GetLibraryVersion: %s\n", err->message);
        g_error_free(err);
        g_object_unref(g_proxy);
        dbus_g_connection_unref(connection);
        return 1;
    }
    printf("libMirage version used by daemon: %s\n", version);
    g_free(version);

    gint iface_version;
    err = NULL;
    dbus_g_proxy_call(g_proxy, "GetDaemonInterfaceVersion", &err,
                      G_TYPE_INVALID, G_TYPE_INT, &iface_version, G_TYPE_INVALID);

    if (err) {
        printf("Error calling GetDaemonInterfaceVersion: %s\n", err->message);
        g_error_free(err);
        g_object_unref(g_proxy);
        dbus_g_connection_unref(connection);
        return 1;
    }

    if (iface_version != 1) {
        printf("Unknown Daemon interface version (%d)\n", iface_version);
        return 1;
    }

    setup_tray();
    connect_signals();
    gtk_main();

    disconnect_signals();
    g_object_unref(g_proxy);
    dbus_g_connection_unref(connection);

    return 0;
}
